from typing import Optional

from ..base import Base
from ..mode import Mode
from .workflow import DBCalculation, Workflow


class pKaMicrostate(Base):
    atom_index: int
    structures: list[DBCalculation] = []
    deltaG: float
    pka: float


class pKaWorkflow(Workflow):
    mode: Mode = Mode.CAREFUL

    pka_range: tuple[float, float] = (2, 12)
    deprotonate_elements: list[int] = [7, 8, 16]
    deprotonate_atoms: list[int] = []
    protonate_elements: list[int] = [7]
    protonate_atoms: list[int] = []

    reasonableness_buffer: float = 5

    structures: list[DBCalculation] = []
    conjugate_acids: list[pKaMicrostate] = []
    conjugate_bases: list[pKaMicrostate] = []
    strongest_acid: Optional[float] = None
    strongest_base: Optional[float] = None



from typing import Any, Optional

from ..mode import Mode
from ..solvent import Solvent
from .workflow import Workflow


class RedoxPotentialWorkflow(Workflow):
    mode: Mode = Mode.RAPID
    solvent: Solvent = Solvent.ACETONITRILE
    reduction: bool = True
    oxidation: bool = True

    # legacy values - remove in future release!
    redox_type: Optional[str] = None
    redox_potential: Optional[float] = None

    # uuids
    neutral_molecule: Optional[str] = None
    anion_molecule: Optional[str] = None
    cation_molecule: Optional[str] = None

    reduction_potential: Optional[float] = None
    oxidation_potential: Optional[float] = None

    def model_post_init(self, __context: Any) -> None:
        """Keep back-compatible with old schema."""
        if self.redox_type == "oxidation":
            self.oxidation = True
            self.reduction = False
            if self.oxidation_potential is None:
                self.oxidation_potential = self.redox_potential
        elif self.redox_type == "reduction":
            self.oxidation = False
            self.reduction = True
            if self.reduction_potential is None:
                self.reduction_potential = self.redox_potential



from .base import LowercaseStrEnum


class Method(LowercaseStrEnum):
    HARTREE_FOCK = "hf"
    HF3C = "hf_3c"

    PBE = "pbe"
    B973C = "b97_3c"
    R2SCAN = "r2scan"
    R2SCAN3C = "r2scan_3c"
    TPSS = "tpss"
    M06L = "m06l"

    PBE0 = "pbe0"
    B3LYP = "b3lyp"
    TPSSH = "tpssh"
    M06 = "m06"
    M062X = "m062x"

    CAMB3LYP = "camb3lyp"
    WB97XD3 = "wb97x_d3"
    WB97XV = "wb97x_v"
    WB97MV = "wb97m_v"
    WB97MD3BJ = "wb97m_d3bj"
    WB97X3C = "wb97x_3c"

    DSDBLYPD3BJ = "dsd_blyp_d3bj"

    AIMNET2_WB97MD3 = "aimnet2_wb97md3"

    # GFN0_XTB = "gfn1_xtb"
    GFN1_XTB = "gfn1_xtb"
    GFN2_XTB = "gfn2_xtb"
    GFN_FF = "gfn_ff"

    # this was going to be removed, but Jonathon wrote such a nice basis set test... it's off the front end.
    BP86 = "bp86"



import pydantic

from .base import Base, LowercaseStrEnum


class DIISStrategy(LowercaseStrEnum):
    # regular pulay DIIS
    DIIS = "diis"

    # Hu/Yang JCP 2010 - ADIIS
    ADIIS = "adiis"

    # first ADIIS, then DIIS
    ADIIS_DIIS = "adiis_diis"


class DIISSettings(Base):
    strategy: DIISStrategy = DIISStrategy.ADIIS_DIIS
    subspace_size: pydantic.PositiveInt = 12

    # if it's a hybrid strategy, where do we transition?
    adiis_diis_blend_start: pydantic.PositiveFloat = 1e-1
    adiis_diis_blend_stop: pydantic.PositiveFloat = 1e-4



import pydantic
from pydantic import PositiveFloat, PositiveInt

try:
    from typing import Optional, Self
except ImportError:
    from typing_extensions import Optional, Self

from .base import Base


class BasisSetOverride(Base):
    name: str
    atomic_numbers: Optional[list[PositiveInt]] = None
    atoms: Optional[list[PositiveInt]] = None  # 1-indexed

    @pydantic.model_validator(mode="after")
    def check_override(self) -> Self:
        # ^ is xor
        assert (self.atomic_numbers is not None) ^ (self.atoms is not None), "Exactly one of ``atomic_numbers`` or ``atoms`` must be specified!"
        return self


class BasisSet(Base):
    name: str

    # do we want to override the default basis set for specific atoms or elements?
    overrides: Optional[list[BasisSetOverride]] = []

    # value below which a basis function can be ignored
    # (for improving DFT grid calcs, as per Stratmann/Scuseria/Frisch CPL 1996)
    # this shouldn't really need to be modified...
    cutoff_threshold: PositiveFloat = 1e-10



from .base import Base, LowercaseStrEnum


class ConstraintType(LowercaseStrEnum):
    """Different sorts of constraints."""

    BOND = "bond"
    ANGLE = "angle"
    DIHEDRAL = "dihedral"


class Constraint(Base):
    """Represents a single constraint."""

    constraint_type: ConstraintType
    atoms: list[int]  # 1-indexed



# ruff: noqa: I001

from .calculation import *
from .molecule import *
from .workflows import *

from .scf_settings import *
from .int_settings import *
from .opt_settings import *
from .diis_settings import *
from .thermochem_settings import *
from .grid_settings import *
from .settings import *

from .method import *
from .basis_set import *
from .task import *
from .correction import *
from .solvent import *
from .mode import *
from .status import *
from .constraint import *
from .message import *



from .workflow import Workflow

Descriptors = dict[str, dict[str, float] | tuple[float | None, ...] | float]


class DescriptorsWorkflow(Workflow):
    # uuid of optimization
    optimization: str | None = None

    descriptors: Descriptors | None = None



from pydantic import Field, PositiveFloat, PositiveInt

from .base import Base
from .constraint import Constraint


class OptimizationSettings(Base):
    max_steps: PositiveInt = 250
    transition_state: bool = False

    # when are we converged?
    max_gradient_threshold: PositiveFloat = 4.5e-4
    rms_gradient_threshold: PositiveFloat = 3.0e-4
    energy_threshold: PositiveFloat = 1e-6

    constraints: list[Constraint] = Field(default_factory=list)



from typing import Optional

from .workflow import Workflow


class FukuiIndexWorkflow(Workflow):
    # uuid of optimization
    optimization: Optional[str] = None

    global_electrophilicity_index: Optional[float] = None
    fukui_positive: Optional[list[float]] = None
    fukui_negative: Optional[list[float]] = None
    fukui_zero: Optional[list[float]] = None



import pydantic
from pydantic import NonNegativeInt, PositiveInt

try:
    from typing import Optional, Self
except ImportError:
    from typing_extensions import Optional, Self

from .base import Base


class VibrationalMode(Base):
    frequency: float
    reduced_mass: float
    force_constant: float
    displacements: list[list[float]]


class Atom(Base):
    atomic_number: NonNegativeInt
    position: list[float]


class Molecule(Base):
    charge: int
    multiplicity: PositiveInt
    atoms: list[Atom]

    energy: Optional[float] = None
    scf_iterations: Optional[NonNegativeInt] = None
    scf_completed: Optional[bool] = None
    elapsed: Optional[float] = None

    homo_lumo_gap: Optional[float] = None  # in eV

    gradient: Optional[list[list[float]]] = None  # Hartree/Bohr

    mulliken_charges: Optional[list[float]] = None
    mulliken_spin_densities: Optional[list[float]] = None
    dipole: Optional[list[float]] = None

    vibrational_modes: Optional[list[VibrationalMode]] = None

    zero_point_energy: Optional[float] = None
    thermal_energy_corr: Optional[float] = None
    thermal_enthalpy_corr: Optional[float] = None
    thermal_free_energy_corr: Optional[float] = None

    @property
    def coordinates(self) -> list[list[float]]:
        return [a.position for a in self.atoms]

    @property
    def atomic_numbers(self) -> list[NonNegativeInt]:
        return [a.atomic_number for a in self.atoms]

    @property
    def sum_energy_zpe(self) -> Optional[float]:
        if (self.energy is None) or (self.zero_point_energy is None):
            return None
        return self.energy + self.zero_point_energy

    @property
    def sum_energy_thermal_corr(self) -> Optional[float]:
        if (self.energy is None) or (self.thermal_energy_corr is None):
            return None
        return self.energy + self.thermal_energy_corr

    @property
    def sum_energy_enthalpy(self) -> Optional[float]:
        if (self.energy is None) or (self.thermal_enthalpy_corr is None):
            return None
        return self.energy + self.thermal_enthalpy_corr

    @property
    def sum_energy_free_energy(self) -> Optional[float]:
        if (self.energy is None) or (self.thermal_free_energy_corr is None):
            return None
        return self.energy + self.thermal_free_energy_corr

    @pydantic.model_validator(mode="after")
    def check_electron_sanity(self) -> Self:
        num_electrons = sum(self.atomic_numbers) - self.charge
        num_unpaired_electrons = self.multiplicity - 1
        if (num_electrons - num_unpaired_electrons) % 2 != 0:
            raise ValueError(
                f"The combination of {num_electrons} electrons, charge {self.charge}, and multiplicity {self.multiplicity} is impossible. "
                "Double-check the charge and multiplicity values given and verify that they are correct."
            )

        return self



from typing import Optional

import numpy as np

from ..base import Base
from ..molecule import Molecule
from ..settings import Settings
from .workflow import Workflow


class ScanPoint(Base):
    index: int
    molecule: Molecule
    energy: Optional[float] = None
    uuid: Optional[str] = None


class ScanSettings(Base):
    type: str  # "bond", "angle", "dihedral" - make Enum later
    atoms: list[int]
    start: float
    stop: float
    num: int

    def vals(self) -> np.ndarray:
        return np.linspace(self.start, self.stop, self.num)

    class Config:
        from_attributes = True


class ScanWorkflow(Workflow):
    scan_settings: ScanSettings
    calc_settings: Settings
    calc_engine: str

    # uuids of scan points
    scan_points: list[str] = []
